import math
import uuid

from .mutex import mutex
from .topology import Topology


class Gossip:
    FANOUT = 5

    def __init__(self, transport, fanout=None, node_id=None):
        self.node_id = node_id or uuid.uuid4().hex
        self.transport = transport

        self.topology = Topology(self.node_id, self.transport.addr)

        self._fanout = fanout or self.FANOUT

    @property
    def fanout(self):
        return min(self._fanout, len(self.topology))

    @property
    def fanout_cycles(self):
        if self.fanout == 0:
            return 0

        if self.fanout == 1:
            return 1

        return math.ceil(math.log(len(self.topology), self.fanout))

    async def send(self, message, node_id):
        message["metadata"]["dst"] = node_id
        self.topology.set_route(message)
        addr = self.topology.get_addr(node_id)
        await self.transport.send(message, addr)

    async def gossip(self, message):
        gossip_id = message["metadata"]["gossip"] = message["metadata"].get(
            "gossip", uuid.uuid4().hex
        )

        fanout_ignore = set([self.node_id])
        fanout_ignore.update([r[1] for r in message["metadata"].get("route", [])])

        @mutex(gossip_id, owner=self.gossip)
        async def fanout():
            cycle = 0
            while cycle < self.fanout_cycles:
                fanout_nodes = self.topology.sample(self.fanout, ignore=fanout_ignore)
                for fanout_node in fanout_nodes:
                    await self.send(message, fanout_node)
                fanout_ignore.update(fanout_nodes)
                cycle += 1

        await fanout()

    async def recv(self):
        while True:
            message, addr = await self.transport.recv()
            message["metadata"]["route"][-1].append(list(addr))
            self.topology.set_route(message)
            self.topology.update_routes(message["metadata"]["route"])

            node_dst = message["metadata"].get("dst", self.node_id)
            node_ack = message["metadata"].get("ack")

            if node_dst == self.node_id:
                if node_ack and node_ack != self.node_id:
                    await self.send(message, node_ack)
            else:
                await self.send(message, node_dst)

            if "gossip" in message["metadata"]:
                await self.gossip(message)

            yield message

    async def close(self):
        self.transport.close()



import collections

# FIXME: use time-based mutexes
MUTEX_CAPACITY = 1024
MUTEX = collections.defaultdict(lambda: collections.deque(maxlen=MUTEX_CAPACITY))


def mutex(mutex_id, owner=None):
    def decorator(func):
        mutexes = MUTEX[owner or func]

        async def wrapper(*args, **kwargs):
            if mutex_id in mutexes:
                return

            mutexes.append(mutex_id)
            return await func(*args, **kwargs)

        return wrapper

    return decorator



import asyncio
import collections
import fnmatch
import itertools

from .channel import Channel
from .errors import print_exception
from .gossip import Gossip


class Callback:
    def __init__(self, topic, func, loop: asyncio.AbstractEventLoop):
        self._loop = loop

        self.topic = topic
        self.func = func
        self.chan = Channel(loop=self._loop)

        self.task = self._loop.create_task(self())
        self.task.add_done_callback(print_exception)

    async def __call__(self):
        while True:
            message = await self.chan.recv()
            await self.func(message)

    async def cancel(self):
        await self.chan.close()
        self.task.cancel()


class Broker:
    def __init__(self, gossip: Gossip, loop: asyncio.AbstractEventLoop):
        self._loop = loop

        self.gossip = gossip
        self.callbacks = collections.defaultdict(list)

    async def listen(self):
        """Connect to the gossip network and start receiving messages."""
        async for message in self.gossip.recv():
            # FIXME: make messages idempotent (prevent duplicate processing)

            if "topic" not in message["metadata"]:
                continue

            topics = list(self.callbacks.keys())

            for topic in topics:
                if fnmatch.fnmatch(message["metadata"]["topic"], topic):
                    for callback in self.callbacks[topic]:
                        await callback.chan.send(message)

            for topic in topics:
                if len(self.callbacks[topic]) == 0:
                    del self.callbacks[topic]

    async def close(self):
        """Disconnect from the gossip network and stop receiving messages."""
        callbacks = [cb for cb in itertools.chain(*self.callbacks.values())]
        await asyncio.gather(*[cb.cancel() for cb in callbacks], return_exceptions=True)
        await self.gossip.close()

    def subscribe(self, topic, callback):
        """Subscribe to a topic and register a callback."""
        callback = Callback(topic, callback, loop=self._loop)
        self.callbacks[topic].append(callback)
        return callback

    async def unsubscribe(self, callback):
        """Unsubscribe from a topic and unregister a callback."""
        await callback.cancel()
        self.callbacks[callback.topic].remove(callback)

    async def publish(self, topic, message, nodes=None):
        """Publish a message to a topic."""
        # FIXME: make messages idempotent (prevent duplicate processing)
        # FIXME: allow sending to specific nodes
        message["metadata"]["topic"] = topic
        if nodes:
            for node in nodes:
                node = self.gossip.topology[node]
                await self.gossip.send(message, node["node_id"])
        else:
            await self.gossip.gossip(message)



import asyncio
import json

import networkx as nx

from aiogossip.peer import Peer

loop = asyncio.get_event_loop()

peer0 = Peer(loop=loop, node_id="p0")
peer1 = Peer(loop=loop, node_id="p1")
peer1.connect([peer0.node])
peer2 = Peer(loop=loop, node_id="p2")
peer2.connect([peer1.node])
peer3 = Peer(loop=loop, node_id="p3")
peer3.connect([peer2.node, peer1.node])


@peer3.subscribe("*")
async def recv(message):
    print("RECV", message, "\n")
    if message.get("message") == "RELAY":
        print("RELAY", message["metadata"]["route"], "\n")


async def main():
    message = {"metadata": {}, "message": "RELAY"}
    await asyncio.sleep(0.1)  # wait for connections to be established
    await peer0.publish("test", message, nodes=[peer3.node_id])


if __name__ == "__main__":
    loop.create_task(main())
    peer0.run_forever()
    loop.run_until_complete(peer0.disconnect())
    loop.run_until_complete(peer1.disconnect())
    loop.run_until_complete(peer2.disconnect())
    loop.run_until_complete(peer3.disconnect())
    loop.close()

    print(json.dumps(nx.to_dict_of_dicts(peer1.gossip.topology.graph)))
    print(json.dumps(nx.to_dict_of_dicts(peer2.gossip.topology.graph)))
    print(json.dumps(nx.to_dict_of_dicts(peer3.gossip.topology.graph)))



import asyncio
import logging
import os
import socket
import sys

from . import codec

logger = logging.getLogger(__name__)
logger.addHandler(logging.StreamHandler(sys.stdout))
if os.environ.get("DEBUG"):
    logger.setLevel(logging.DEBUG)  # pragma: no cover


class Transport:
    PAYLOAD_SIZE = 4096

    def __init__(self, bind, loop: asyncio.AbstractEventLoop):
        self._loop = loop

        self.sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.sock.bind(bind)
        self.sock.setblocking(False)

        self.rx_bytes = 0
        self.rx_packets = 0
        self.tx_bytes = 0
        self.tx_packets = 0

    @property
    def addr(self):
        return self.sock.getsockname()

    async def send(self, message, addr):
        message = codec.encode(message)
        if len(message) > self.PAYLOAD_SIZE:
            raise ValueError(
                f"Message size exceeds payload size of {self.PAYLOAD_SIZE} bytes"
            )
        await self._loop.sock_sendto(self.sock, message, addr)
        logger.debug(f"DEBUG: {self.addr[1]} > {addr[1]} send: {message}\n")

        self.tx_packets += 1
        self.tx_bytes += len(message)

    async def recv(self):
        message, addr = await self._loop.sock_recvfrom(self.sock, self.PAYLOAD_SIZE)
        logger.debug(f"DEBUG: {self.addr[1]} < {addr[1]} recv: {message}\n")

        self.rx_packets += 1
        self.rx_bytes += len(message)
        message = codec.decode(message)
        return message, addr

    def close(self):
        self.sock.close()



import asyncio
import uuid

from .broker import Broker
from .errors import print_exception
from .gossip import Gossip
from .transport import Transport


class Peer:
    def __init__(
        self,
        loop: asyncio.AbstractEventLoop,
        host="0.0.0.0",
        port=0,
        fanout=None,
        node_id=None,
    ):
        self._loop = loop

        self.node_id = node_id or uuid.uuid4().hex
        # FIXME: should be lazy
        self.transport = Transport((host, port), loop=self._loop)
        self.gossip = Gossip(self.transport, fanout=fanout, node_id=self.node_id)
        self.broker = Broker(self.gossip, loop=self._loop)

        self.task = self._loop.create_task(self.broker.listen())
        self.task.add_done_callback(print_exception)

    @property
    def node(self):
        return self.gossip.topology.node

    async def _connect(self):
        message = {"metadata": {}}
        await self.publish("connect", message, ack=True)

    def connect(self, nodes):
        self.gossip.topology.add(nodes)
        task = self._loop.create_task(self._connect())
        task.add_done_callback(print_exception)

    async def disconnect(self):
        await self.broker.close()
        self.task.cancel()
        await asyncio.gather(self.task, return_exceptions=True)

    async def publish(self, topic, message, nodes=None, ack=False):
        if ack:
            message["metadata"]["ack"] = self.node_id
        await self.broker.publish(topic, message, nodes=nodes)

    def subscribe(self, topic):
        def decorator(callback):
            callback = self.broker.subscribe(topic, callback)
            return callback

        return decorator

    def run_forever(self):  # pragma: no cover
        try:
            self._loop.run_forever()
        except KeyboardInterrupt:
            pass
        finally:
            self._loop.run_until_complete(self.disconnect())



import asyncio
import json

import networkx as nx

from aiogossip.peer import Peer

loop = asyncio.get_event_loop()

peer1 = Peer(loop=loop, node_id="p1")


@peer1.subscribe("test")
async def handler1(message):
    print("handler1", message, "\n")
    reply = {"message": "bar", "metadata": {}}
    await peer1.publish("test", reply)


peer2 = Peer(loop=loop, node_id="p2")
peer2.connect([peer1.node])


@peer2.subscribe("test")
async def handler2(message):
    print("handler2", message, "\n")


peer3 = Peer(loop=loop, node_id="p3")
peer3.connect([peer1.node])


@peer3.subscribe("*")
async def handler3(message):
    print("handler3", message, "\n")


peer4 = Peer(loop=loop, node_id="p4")
peer4.connect([peer3.node])


@peer4.subscribe("*")
async def handler4(message):
    print("handler4", message, "\n")


async def main():
    message = {"message": "foo", "metadata": {}}
    await asyncio.sleep(0.1)  # wait for connections to be established
    await peer2.publish("test", message)


if __name__ == "__main__":
    loop.create_task(main())
    peer1.run_forever()
    loop.run_until_complete(peer1.disconnect())
    loop.run_until_complete(peer2.disconnect())
    loop.run_until_complete(peer3.disconnect())
    loop.run_until_complete(peer4.disconnect())
    loop.close()

    print(json.dumps(nx.to_dict_of_dicts(peer1.gossip.topology.graph)))
    print(json.dumps(nx.to_dict_of_dicts(peer2.gossip.topology.graph)))
    print(json.dumps(nx.to_dict_of_dicts(peer3.gossip.topology.graph)))
    print(json.dumps(nx.to_dict_of_dicts(peer4.gossip.topology.graph)))



import random
import time

import networkx as nx


class Topology:
    def __init__(self, node_id, node_addr):
        self.graph = nx.DiGraph(node_id=node_id, node_addr=node_addr)
        self.add([self.node])

    # Node #
    @property
    def node_id(self):
        return self.graph.graph["node_id"]

    @property
    def node_addr(self):
        return self.graph.graph["node_addr"]

    @property
    def node(self):
        return {"node_id": self.node_id, "node_addr": tuple(self.node_addr)}

    # Topology #
    def add(self, nodes):
        for node in nodes:
            self.graph.add_node(node["node_id"], **node)

            if node["node_id"] == self.node_id:
                continue

            src = self.graph.nodes[self.node_id]
            dst = self.graph.nodes[node["node_id"]]
            attrs = {
                "src": tuple(src["node_addr"]),
                "dst": tuple(dst["node_addr"]),
                "latency": 0,
            }
            self.graph.add_edge(src["node_id"], dst["node_id"], **attrs)

    def sample(self, k, ignore=None):
        nodes = [e[1] for e in self.graph.out_edges(self.node_id)]
        if ignore:
            nodes = list(set(nodes) - set(ignore))
        k = min(k, len(nodes))
        return random.sample(nodes, k)

    def __getitem__(self, node_id):
        return self.graph.nodes[node_id]

    def __len__(self):
        return len(self.graph)

    def __iter__(self):
        return iter(self.graph.nodes)

    def __contains__(self, node_id):
        return node_id in self.graph

    # Route #
    @property
    def route(self):
        return [time.time_ns(), self.node_id, self.node_addr]

    def set_route(self, message):
        route = message["metadata"].get("route", [self.route])
        if route[-1][1] != self.graph.graph["node_id"]:
            route.append(self.route)
        message["metadata"]["route"] = route
        return message

    def update_routes(self, routes):
        def add(src, dst):
            attrs = {
                "src": tuple(src[-1]),
                "dst": tuple(dst[-1]),
                "latency": abs(src[0] - dst[0]),
            }
            self.graph.add_node(src[1], node_id=src[1], node_addr=tuple(src[-1]))
            self.graph.add_node(dst[1], node_id=dst[1], node_addr=tuple(dst[-1]))
            self.graph.add_edge(src[1], dst[1], **attrs)

        connections = [(routes[i], routes[i + 1]) for i in range(len(routes) - 1)]
        if not connections:
            raise ValueError("route must contain at least two nodes")

        dst, src = connections[-1]
        add(src, dst)
        for src, dst in connections:
            add(src, dst)

    # Addr #
    def get_addr(self, node_id):
        path = nx.shortest_path(self.graph.to_undirected(), self.node_id, node_id)
        return self.graph.edges[path[0], path[1]]["dst"]
